/* ========================== */
/* = SERVICE - Media player = */
/* ========================== */
lm.factory('serviceMediaPlayer', ['$rootScope', '$http', 'serviceNotification', 'serviceLimboMedia', 'serviceFullscreen', 'serviceSidebar', function($rootScope, $http, serviceNotification, serviceLimboMedia, serviceFullscreen, serviceSidebar) { 
	var service = new Object();
	
	/* =========== */
	/* = STATICS = */
	/* =========== */
	service.layoutMap = {'P360': '360p', 'P720': '720p', 'P1080': '1080p', 'P2160': '4K'};
	service.supportedVideo = [];	// effectively static: filled on first binding
	service.supportedAudio = [];	// effectively static: filled on first binding

	/* ============ */
	/* = SETTINGS = */
	/* ============ */
	service.settings = {videoApproach: 'DEFAULT'};
	
	/* ============ */
	/* = BINDINGS = */
	/* ============ */
	service.binding = {
		bound: false,		// successfully bound?
		video: null,		// video element 
		frame: null,		// frame getting fullscreened
	};

	/* Destroy to unbind eventhandlers. */
	service.unbind = function() {
		console.log("ServiceMediaPlayer - Destroying Player.");
		// TODO implement unbinding
	}
	
	/* Must be called when dom-elements for video and fullscreen-frame are available. Required for MediaPlayerService to work cause then we can check supported formats and bind playback-event-handlers. */
	service.bind = function(videoElement, frameElement) {
		service.binding.video = videoElement;
		service.binding.frame = frameElement;
		
		// Prepare subtites with captionator:
		service.binding.video._captionatorOptions = new Object();
		service.binding.video._captionatorOptions.appendCueCanvasTo = service.binding.frame;
		
		// Check supported video and audio containers and codecs:
		if(service.supportedVideo.length == 0 && service.supportedAudio.length == 0) {
			if(service.binding.video.canPlayType("video/webm").match(/maybe|probably/i)) service.supportedVideo.push("WEBM");
			if(service.binding.video.canPlayType("video/mp4").match(/maybe|probably/i)) service.supportedVideo.push("MP4");
			if(service.binding.video.canPlayType("audio/webm").match(/maybe|probably/i)) service.supportedAudio.push("WEBM_VORBIS");
			if(service.binding.video.canPlayType("audio/mp3").match(/maybe|probably/i)) service.supportedAudio.push("MP3");
			if(service.binding.video.canPlayType("audio/ogg").match(/maybe|probably/i)) service.supportedAudio.push("OGG_VORBIS");
			// console.log("Supported Video: " + service.supportedVideo);
			// console.log("Supported Audio: " + service.supportedAudio);
		}
		
		// Bind listener
		service.binding.video.addEventListener('timeupdate',service.refreshPosition);
		service.binding.video.addEventListener('playing',service.refreshPlaying);
		service.binding.video.addEventListener('pause',service.refreshPlaying);
		service.binding.video.addEventListener('ended',service.refreshEnded);	
		
		// Initial volume
		service.binding.video.volume = service.state.volume; 
		
		// Ready to rumble
		service.binding.bound = true;
	}
	
	/* ========= */
	/* = STATE = */
	/* ========= */	
	
	service.state = {
		file: null,			// current file
		playing: false,
		offset: 0,			// transcoding offset position (seconds)
		position: 0,		// aggregated playback position (offset + position in stream) (seconds)
		volume: 0.5,		// volume (range 0 - 1)
		format: null,
		formats: [],
		astream: null,
		astreams: [],
		sstream: null,
		sstreams: [{idx:-1, type:'SUB', info:'subtitles off', language:'OFF', codec:null}],
	};
	
	service.resetState = function(hardreset) {
		if(service.binding.bound) {
			service.binding.video.pause();
			try { service.binding.video.currentTime = 0; } catch(error) {};
		}

		service.state.file = null;
		service.state.playing = false;
		service.state.offset = 0;
		service.state.position = 0;
		service.state.format = null;
		service.state.formats = [];
		service.state.astream = null;
		service.state.astreams = [];
		service.state.sstream = null;
		service.state.sstreams.splice(1,service.state.sstreams.length - 1);
		if(hardreset) {
			service.state.volume = 0.5;
		}
		service.subs.reset();
	};	
	
	/* ============ */
	/* = PLAYLIST = */
	/* ============ */	
	
	service.pl = {
		playlist: {id: 0, name: 'last', items: []},
		idx: 0,
		random: false,
		repeat: false,
		repeatOne: false
	};

	service.resetPlaylist = function() {
		service.pl.playlist = {id: 0, name: 'last', items: []};
		service.pl.idx = 0;
		service.pl.random = false;
		service.pl.repeat = false;
		service.pl.repeatOne = false;
	};

	/* ==================== */
	/* = NATIVE CALLBACKS = */
	/* ==================== */

	/* Update own state when player changes position */ 
	service.refreshPosition = function(dingens) {
		var src = this; // video or audio element. alt: dingens.srcElement. src == service.elAudioPlayer
		$rootScope.$apply(function() {
			service.state.position = Number(service.state.offset) + Number(src.currentTime);
		});
		
		service.subs.refreshState();
	};

	/* Update own state when player changes play/pause/... */
	service.refreshPlaying = function() {
		$rootScope.$apply(function() {
			service.state.playing = !(service.binding.video.paused);
		});
	};
	
	/* Handle playback finished (Next track?) */
	service.refreshEnded = function() {
		$rootScope.$apply(function() {
			if(service.pl.repeat && service.pl.repeatOne) service.intPlayAgain();
			else service.playNext();
		});
	};
	
	/* ============ */
	/* = INTERNAL = */
	/* ============ */

	/* Play some media file. Load details (formats, ...) when not available. */
	service.intPlayFile = function(valueFile) {
		/* Loading files with details (formats, ...) initially is expensive. Therefore we lazy-load details here as soon as required. */
		if(valueFile.formats === undefined || valueFile.formats.length == 0) {
			$http.get(lmBase.url + 'f/file/' + valueFile.id).then(
				function success(resp) {
					valueFile.formats = resp.data.formats;
					service.intPlayFileWithDetails(valueFile);
				},
				serviceLimboMedia.handleHttpErrorNew
			);
		} else {
			service.intPlayFileWithDetails(valueFile);
		}
	};
	
	/* Play some media file. Call as soon as file details (formats, ...) are available. */
	service.intPlayFileWithDetails = function(valueFile) {
		service.resetState();
		service.state.file = valueFile;
		
		if(service.state.file.type == 'AUDIO') {
			service.playAudio(valueFile);
		}
		else if(service.state.file.type == 'VIDEO') {
			service.playVideo(valueFile);
		}
	};	
	
	service.playAudio = function(valueFile) {
		var supportedStandard = null;
		for(x in valueFile.formats) {
			var af = valueFile.formats[x];
			if(service.supportedAudio.indexOf(af.standard) != -1) {
				supportedStandard = af.standard;
				break;
			}
		}

		var playurl = lmBase.url + 'f/streamAudio/' + valueFile.id +'/x' + '?standard=' + (supportedStandard == null ? 'WEBM_VORBIS' : supportedStandard);

		service.binding.video.src = playurl;
		service.binding.video.play();
	}
	
	service.playVideo = function(valueFile) {
		/* extract audio and sub streams */
		for(x in valueFile.formats) {
			var format = valueFile.formats[x];
			if(format.original) {
				for(y in format.streams) {
					var stream = format.streams[y];
					if("AUDIO" == stream.type) {
						service.state.astreams.push(stream);
					}
					if("SUB" == stream.type && ('SUBRIP' == stream.codec || 'WEBVTT' == stream.codec)) {
						service.state.sstreams.push(stream);
					}
				}
				if(service.state.astreams.length > 0) {
					service.state.astream = service.state.astreams[0];
				}
			}
		}
		
		
		/* Add available streams, original and supported standards */
		for(x in valueFile.formats) {
			var format = valueFile.formats[x];
			if(format.original || service.supportedVideo.indexOf(format.standard) != -1) {
				service.state.formats.push(format);
			}
		}

		/* Add transcoding streams, only supported standards */
		for(tfi in lmBase.videoFormats) {
			var tf = lmBase.videoFormats[tfi];
			var alreadyThere = false;
			for(x in valueFile.formats) {
				var format = valueFile.formats[x];
				if(!format.original && tf.standard == format.standard && tf.layout == format.layout) {
					alreadyThere = true;
					break;
				}
			}
			if(!alreadyThere && service.supportedVideo.indexOf(tf.standard) != -1) {
				service.state.formats.push(tf);
			}
		}
		
		/* sort */
		service.state.formats.sort(lmBase.compareVideoFormat);
		
		/* add guinames and metainfos */
		for(x in service.state.formats) {
			var format = service.state.formats[x];
			format.guinamesmall = format.original ? 'orig.' : service.layoutMap[format.layout];
			format.guiname = (format.layout == null ? 'Custom' : service.layoutMap[format.layout]);
			format.guiname += ' ' + format.mime.substring(6);

			if(format.original) format.guiname += ' (original)';
			else if(format.transcoding) format.guiname += ' (transcode)';
			
			/* add nicesize and urls, and playable */
			format.url = lmBase.url + 'f/streamVideo/' + valueFile.id + '/' + valueFile.urlname;
			if(!format.original) {
				format.url += '.' + format.standard + '?standard=' + format.standard + '&layout=' + format.layout + '&sau=666';
			}
		}
		
		/* select format to play. Approaches: ['DEFAULT','P360','P720','P1080'] */
		if('P360' == service.settings.videoApproach || 'P720' == service.settings.videoApproach || 'P1080' == service.settings.videoApproach) {
			var possibleFormats = service.state.formats.filter(function(valFormat) {
				return service.settings.videoApproach == valFormat.layout;
			});

			if(possibleFormats.length > 0) {
				service.state.format = possibleFormats[0];
			}
		}
		else if(service.state.formats.length > 0) {
			service.state.format = service.state.formats[0];
		}

		service.refreshVideo(0);
	}
	
	/*
	 * changePosition:
	 *  -1 = Keep position
	 *   0 = Reset to start
	 * 1-x = Set this position
	 * not working right now: switching from transcoding to non-transcoding
	 */
	service.refreshVideo = function(changePosition) {
		if(changePosition == -1) {
			if(service.state.format.transcoding) {
				service.state.offset += Math.floor(Number(service.binding.video.currentTime));	
			}
			else {
				service.state.offset = 0;
			}
		}
		else {
			service.state.offset = changePosition;
		}
		
		var playurl = lmBase.url + 'f/streamVideo/' + service.state.file.id + '/x';
		if(!service.state.format.original) {
			playurl += '?standard=' + service.state.format.standard + '&layout=' + service.state.format.layout;
		}
		
		if(service.state.format.transcoding) {
			if(service.state.offset != 0) {
				playurl += "&ss=" + service.state.offset;
			}
			if(service.state.astream != null) {
				playurl += "&sau=" + service.state.astream.idx;
			}
		}
		
		service.binding.video.src = playurl;
		service.binding.video.play();
		
		service.subs.refreshTrack();
	};	

	/* ===================== */
	/* = INTERNAL PLAYLIST = */
	/* ===================== */
	
	/* Play item at playlist position. */
	service.intPlayIndex = function(valueIndex) {
		if(service.pl.playlist.items.length == 0 || valueIndex < 0) {
			return;
		}

		if(valueIndex >= service.pl.playlist.items.length) {
			if(service.pl.repeat && !service.pl.repeatOne) valueIndex = 0;
			else return;
		}

		service.pl.idx = valueIndex;

		var fileToPlay = service.pl.playlist.items[valueIndex];
		service.intPlayFile(fileToPlay);
	};
	
	service.intPlayRandom = function() {
		service.intPlayIndex(Math.floor(Math.random() * service.pl.playlist.items.length));
	};
	
	service.intPlayAgain = function() {
		service.intPlayIndex(service.pl.idx);
	}
	
	
	
	
	/* ===================== */
	/* = INTERNAL SUBTITLE = */
	/* ===================== */

	trackObject = null;	// Subtitle track object. Must be global for unknown captionator reasons.

	service.subs = {
		reset: function() {
			if(trackObject) {
				trackObject.mode = captionator.TextTrack.OFF;
				captionator.rebuildCaptions(service.binding.video);
				trackObject = null;			
			}
		},
		
		/* Call whenever position changes to refresh text. */	
		refreshState: function() {
			if(trackObject) {
				try {
					trackObject.activeCues.refreshCues.apply(trackObject.activeCues);
				} catch(error) {}
				captionator.rebuildCaptions(service.binding.video);
			}
		},
		
		/* Call whenever selected subtitle track changes. */
		refreshTrack: function() {
			if(service.state.sstream) {
				var earl = lmBase.url + 'f/streamVideoSub/' + service.state.file.id + '/' + service.state.sstream.idx;
				
				if(service.state.offset != 0) {
					earl += "?ss=" + service.state.offset;
				}
				
				trackObject = new captionator.TextTrack(captionator.generateID(),'subtitles','blub','de',earl,null);
				trackObject.videoNode = service.binding.video;
				trackObject.mode = captionator.TextTrack.SHOWING;
			}
			else {
				service.subs.reset();
			}
			
			/*		
				window.addEventListener("resize", function(eventData) {
					videoElement._captionator_dirtyBit = true; // mark video as dirty, force captionator to rerender captions
					captionator.rebuildCaptions(videoElement);
				},false);
			*/
		}
	};

	
	/* ============================ */
	/* = EXTERNAL COMMANDS PLAYER = */
	/* ============================ */	
	
	/* Set volume. Input parameter range: number between 0.00 - 1.00. */
	service.changeVolume = function(valueNewVolume) {
		service.state.volume = valueNewVolume;
		if(service.binding.bound) service.binding.video.volume = service.state.volume;
	}
	
	/* Set position. Input parameter range: number between 0.00 - 1.00. */
	service.changePosition = function(valueNewPosition) {
		if(!service.state.file || !service.binding.bound) return;

		var newpos = valueNewPosition * service.state.file.duration;
		
		if('VIDEO' == service.state.file.type && service.state.format.transcoding) {
			service.refreshVideo(Math.round(newpos));
		}
		else {
			// TODO Support offset for transcoding audio
			service.binding.video.currentTime = newpos;
		}
	}
	
	/* Play if paused, pause if playing. */
	service.playPause = function() {
		if(service.state.file == null || !service.binding.bound) return;
		if (service.binding.video.paused || service.binding.video.ended) service.binding.video.play();
		else service.binding.video.pause();
	};	
	
	/* Stop player and reset state. */
	service.stop = function() {
		service.resetState();
	};
	
	/* Reset current track. */
	service.reset = function() {
		if(service.state.file == null || !service.binding.bound) return;
		service.intPlayFile(service.state.file);
	};
	
	/* Change video stream. */
	service.changeFormat = function(format) {
		service.state.format = format;
		service.refreshVideo(-1);
	};
	
	/* Change audio stream. */
	service.changeAudioStream = function(as) {
		service.state.astream = as;
		service.refreshVideo(-1);
	};
	
	/* Change subtitle stream. */
	service.changeSubStream = function(ss) {
		service.state.sstream = ss.idx == -1 ? null : ss;
		service.subs.refreshTrack();
	};
	
	service.toggleFullscreen = function() {
		serviceFullscreen.onToggleFullscreen(service.binding.frame);
	};

	
	/* ============================== */
	/* = EXTERNAL COMMANDS PLAYLIST = */
	/* ============================== */
	
	service.playIndex = function(valIndex) {
		service.intPlayIndex(valIndex);
	}
	
	service.playPrev = function() {
		if(service.pl.random) service.intPlayRandom();
		else service.intPlayIndex(service.pl.idx - 1);
	};
	
	service.playNext = function() {
		if(service.pl.random) service.intPlayRandom();
		else service.intPlayIndex(service.pl.idx + 1);
	};
	
	service.playFile = function(valueFile) {
		service.pl.playlist.items = [valueFile];
		service.intPlayIndex(0);
		service.playlistOnChange();
		serviceSidebar.setShow(true);
	};

	service.playFiles = function(valueFiles) {
		service.pl.playlist.items = new Array();
		for(x in valueFiles) {
			service.pl.playlist.items.push(valueFiles[x]);
		}
		service.intPlayIndex(0);
		service.playlistOnChange();
		serviceSidebar.setShow(true);
	};

	service.addFile = function(valueFile) {
		service.pl.playlist.items.push(valueFile);
		service.playlistOnChange();
	};
	
	service.addFiles = function(valueFiles) {
		for(x in valueFiles) {
			service.pl.playlist.items.push(valueFiles[x]);
		}
		service.playlistOnChange();
	};

	service.addPlayFile = function(valueFile) {
		service.pl.playlist.items.push(valueFile);
		service.intPlayIndex(service.pl.playlist.items.length - 1);
		service.playlistOnChange();
		serviceSidebar.setShow(true);
	};	
	
	/* playlist stuff */
	service.playlistsLoadLast = function() {
		$http.get(lmBase.url + 'f/playlists/list').then(
			function success(resp) {
				var found = false;
				for(y in resp.data) {
					var pl = resp.data[y];
					if('last' == pl.name) {
						service.playlistLoad(pl,false);
						found = true;
					}
				}
				if(!found) {
					service.pl.playlist = {id: 0, name: 'last', items: []};
				}
			},
			serviceLimboMedia.handleHttpErrorNew
		);
	};
	
	service.playlistLoad = function(playlist,playit) {
		$http.get(lmBase.url + 'f/playlist/' + playlist.id).then(
			function success(resp) {
				for(x in resp.data.items) {
					lmBase.pimpLmFile(resp.data.items[x]);
				}
				service.pl.playlist = resp.data;
				if(playit) service.intPlayIndex(0);
			},
			serviceLimboMedia.handleHttpErrorNew
		);
	};	

	service.playlistSave = function(valName) {
		var req = {name: valName, items: lmBase.arrayToIdArray(service.pl.playlist.items)};
		$http.post(lmBase.url + 'f/playlist',req).then(
			function success(resp) {
				service.pl.playlist.id = resp.data.id;
				service.pl.playlist.name = resp.data.name;
				serviceNotification.ok('MSG_SAVED');
			},
			serviceLimboMedia.handleHttpErrorNew
		);
	};

	service.playlistSaveLast = function() {
		var req = {name: 'last', items: lmBase.arrayToIdArray(service.pl.playlist.items)};
		$http.post(lmBase.url + 'f/playlist',req).then(
			function success(resp) {},
			serviceLimboMedia.handleHttpErrorNew
		);
	};
	
	service.playlistDelete = function(playlist) {
		$http.delete(lmBase.url + 'f/playlist/' + playlist.id).then(
			function success(resp) {
				service.playlistsLoadLast();
			},
			serviceLimboMedia.handleHttpErrorNew
		);
	};
	
	service.playlistClear = function() {
		service.pl.playlist.items = [];
		service.playlistOnChange();
	};	
	
	service.playlistOnChange = function() {
		if(serviceLimboMedia.hasRole('ROLE_WRITE')) {
			service.playlistSaveLast();
		}
	};

	$rootScope.$on('lmLogin', function() {
		service.settings = serviceLimboMedia.getRemoteSettings();
		if(service.settings.videoApproach == null) service.settings.videoApproach = 'DEFAULT';

		service.playlistsLoadLast();
	});
	
	$rootScope.$on('lmLogout', function() {
		service.resetState(true);
		service.resetPlaylist();
	});
	
	return service;
}]);
