lm.factory('serviceUpload', [ '$rootScope', '$http', 'serviceLimboMedia', 'serviceNotification', 'serviceModal' , function($rootScope, $http, serviceLimboMedia, serviceNotification, serviceModal) {
	var service = new Object();
	
	/* attributes */
	service.queue = [];
	service.failed = [];
	
	/* upload complete listener */
	service.listener = null;
	service.registerListener = function(valueListener) {
		service.listener = valueListener;
	};
	
	/* show upload state */
	service.state = new Object();
	service.state.uploading = false;
	service.state.percent = 0;
	service.state.currentItem = null;
	
	/* status toast */
	service.toastRef = null;
	
	service.toastOn = function() {
		if(!service.toastRef) {
			service.toastRef = serviceNotification.toastShow('lm-ui-toast-upload', 'lm-style-info', true, {});
		}
	};
	
	service.toastOff = function() {
		if(service.toastRef) {
			serviceNotification.toastDone(service.toastRef);
			service.toastRef = null;
		}
	};

	/* monitoring and control */
	service.failedRemove = function(item) {
		var idx = service.failed.indexOf(item);
		if(idx != -1) service.failed.splice(idx, 1);
	};
	
	service.failedRetry = function(item) {
		service.failedRemove(item);
		service.queue.push(item);
		service.toastOn();
		service.next();
	};

	/* functions */
	service.uploadFiles = function(valueFiles, valueURL, valueForm, valueParams) {
		if(!valueFiles || valueFiles.length == 0) {
			return;
		}

		for(x in valueFiles) {
			var item = {
				file: valueFiles[x],
				url: valueURL,
				params: valueParams,
				form: valueForm,
				error: null
			}
			service.queue.push(item);
		}

		/* Show info about uploading */
		service.toastOn();
		
		/* Start upload */
		service.next();
	};
	
	service.next = function() {
		if(service.state.currentItem != null) {
			return;
		}
	
		if(service.queue.length == 0) {
			return;
		}
		
		service.state.currentItem = service.queue.shift();
		service.state.uploading = true;
	
		/* Create XHR */
		var xhr = new XMLHttpRequest();
		
		/* Add listeners */
		xhr.upload.addEventListener("progress", service.uploadProgress, false);
		xhr.addEventListener("load", service.uploadEvent, false);
		xhr.addEventListener("error", service.uploadEvent, false);
		xhr.addEventListener("abort", service.uploadEvent, false);
		
		/* Direct upload (form upload not used anymore) */
		xhr.open("POST", service.state.currentItem.url);
		xhr.setRequestHeader("X-LM-filename", Base64.encode(service.state.currentItem.file.name));
		if(service.state.currentItem.file.lastModified) {
			xhr.setRequestHeader("X-LM-filetsmod", service.state.currentItem.file.lastModified);
		}
		xhr.setRequestHeader("LM-csrf","yeah");
		if(service.state.currentItem.params != null) {
			for(var x in service.state.currentItem.params) {
				xhr.setRequestHeader('X-LM-' + x, service.state.currentItem.params[x]);
			}
		}			
		xhr.send(service.state.currentItem.file);
	};
	
	service.uploadProgress = function(evt) {
		if (evt.lengthComputable) {
			var percentComplete = Math.round(evt.loaded * 100 / evt.total);
			$rootScope.$apply(function() {
				service.state.percent = percentComplete;
			});
		}
	};
	
	service.uploadEvent = function(evt) {
		// this = XMLHttpRequest
		if(200 != this.status) {
			var jsonResponse = {key: 'MSG_UPLOAD_FAIL'};
			try {jsonResponse = angular.fromJson(this.response)} catch (e) { /* failing whenever error isnt json */ }

			$rootScope.$apply(function() {
				serviceLimboMedia.httpErrorTranslate(jsonResponse,function(msg) {
					service.state.currentItem.error = msg;
					service.uploadDone(false);
				});
			});
		}
		else {
			$rootScope.$apply(function() {
				service.uploadDone(true);
			});
		}
	};
	
	service.uploadDone = function(withSuccess) {
		if(!withSuccess) {
			service.failed.push(service.state.currentItem);
		}
		
		service.state.currentItem = null;
		service.state.uploading = false;
		service.state.percent = 0;

		if(service.queue.length == 0) {
			service.toastOff();
			serviceNotification.ok('MSG_UPLOAD_DONE');
			
			if(service.failed.length > 0) {
				serviceModal.show('lm-dialog-upload-fails',{failed: service.failed.splice(0,service.failed.length)}).then(
					function() {},
					function() {}
				);
			}
			
			if(service.listener != null) {
				service.listener();
			}
		}
		service.next();
	};
	
	return service;
}]);
